@@ -1,30 +1,27 @@
 /* Support routines for GNU DIFF.
 
-   Copyright (C) 1988, 1989, 1992, 1993, 1994, 1995, 1998, 2001, 2002
+   Copyright (C) 1988-1989, 1992-1995, 1998, 2001-2002, 2004, 2006, 2009-2011
    Free Software Foundation, Inc.
 
    This file is part of GNU DIFF.
 
-   GNU DIFF is free software; you can redistribute it and/or modify
+   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
+   the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
 
-   GNU DIFF is distributed in the hope that it will be useful,
+   This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; see the file COPYING.
-   If not, write to the Free Software Foundation,
-   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
 #include "diff.h"
 #include <dirname.h>
 #include <error.h>
-#include <quotesys.h>
-#include <regex.h>
+#include <sh-quote.h>
 #include <xalloc.h>
 
 char const pr_program[] = PR_PROGRAM;
@@ -165,7 +162,7 @@ setup_output (char const *name0, char const *name1, bool recursive)
   outfile = 0;
 }
 
-#if HAVE_WORKING_FORK || HAVE_WORKING_VFORK
+#if HAVE_WORKING_FORK
 static pid_t pr_pid;
 #endif
 
@@ -195,13 +192,13 @@ begin_output (void)
 
       /* Make OUTFILE a pipe to a subsidiary `pr'.  */
       {
-#if HAVE_WORKING_FORK || HAVE_WORKING_VFORK
+#if HAVE_WORKING_FORK
 	int pipes[2];
 
 	if (pipe (pipes) != 0)
 	  pfatal_with_name ("pipe");
 
-	pr_pid = vfork ();
+	pr_pid = fork ();
 	if (pr_pid < 0)
 	  pfatal_with_name ("fork");
 
@@ -215,8 +212,8 @@ begin_output (void)
 		close (pipes[0]);
 	      }
 
-	    execl (pr_program, pr_program, "-h", name, 0);
-	    _exit (errno == ENOEXEC ? 126 : 127);
+	    execl (pr_program, pr_program, "-h", name, (char *) 0);
+	    _exit (errno == ENOENT ? 127 : 126);
 	  }
 	else
 	  {
@@ -227,11 +224,11 @@ begin_output (void)
 	  }
 #else
 	char *command = xmalloc (sizeof pr_program - 1 + 7
-				 + quote_system_arg ((char *) 0, name) + 1);
+				 + shell_quote_length (name) + 1);
 	char *p;
 	sprintf (command, "%s -f -h ", pr_program);
 	p = command + sizeof pr_program - 1 + 7;
-	p += quote_system_arg (p, name);
+	p = shell_quote_copy (p, name);
 	*p = 0;
 	errno = 0;
 	outfile = popen (command, "w");
@@ -260,11 +257,11 @@ begin_output (void)
   switch (output_style)
     {
     case OUTPUT_CONTEXT:
-      print_context_header (files, 0);
+      print_context_header (files, false);
       break;
 
     case OUTPUT_UNIFIED:
-      print_context_header (files, 1);
+      print_context_header (files, true);
       break;
 
     default:
@@ -280,11 +277,12 @@ finish_output (void)
 {
   if (outfile != 0 && outfile != stdout)
     {
+      int status;
       int wstatus;
       int werrno = 0;
       if (ferror (outfile))
 	fatal ("write failed");
-#if ! (HAVE_WORKING_FORK || HAVE_WORKING_VFORK)
+#if ! HAVE_WORKING_FORK
       wstatus = pclose (outfile);
       if (wstatus == -1)
 	werrno = errno;
@@ -294,21 +292,19 @@ finish_output (void)
       if (waitpid (pr_pid, &wstatus, 0) < 0)
 	pfatal_with_name ("waitpid");
 #endif
-      if (werrno | wstatus)
-	{
-	  char const *failure_msgid = N_("subsidiary program `%s' failed");
-	  if (! werrno && WIFEXITED (wstatus))
-	    switch (WEXITSTATUS (wstatus))
-	      {
-	      case 126:
-		failure_msgid = N_("subsidiary program `%s' not executable");
-		break;
-	      case 127:
-		failure_msgid = N_("subsidiary program `%s' not found");
-		break;
-	      }
-	  error (EXIT_TROUBLE, werrno, _(failure_msgid), pr_program);
-	}
+      status = (! werrno && WIFEXITED (wstatus)
+		? WEXITSTATUS (wstatus)
+		: INT_MAX);
+      if (status)
+	error (EXIT_TROUBLE, werrno,
+	       _(status == 126
+		 ? "subsidiary program `%s' could not be invoked"
+		 : status == 127
+		 ? "subsidiary program `%s' not found"
+		 : status == INT_MAX
+		 ? "subsidiary program `%s' failed"
+		 : "subsidiary program `%s' failed (exit status %d)"),
+	       pr_program, status);
     }
 
   outfile = 0;
@@ -323,8 +319,8 @@ finish_output (void)
 bool
 lines_differ (char const *s1, char const *s2)
 {
-  register unsigned char const *t1 = (unsigned char const *) s1;
-  register unsigned char const *t2 = (unsigned char const *) s2;
+  register char const *t1 = s1;
+  register char const *t2 = s2;
   size_t column = 0;
 
   while (1)
@@ -339,20 +335,20 @@ lines_differ (char const *s1, char const *s2)
 	    {
 	    case IGNORE_ALL_SPACE:
 	      /* For -w, just skip past any white space.  */
-	      while (ISSPACE (c1) && c1 != '\n') c1 = *t1++;
-	      while (ISSPACE (c2) && c2 != '\n') c2 = *t2++;
+	      while (isspace (c1) && c1 != '\n') c1 = *t1++;
+	      while (isspace (c2) && c2 != '\n') c2 = *t2++;
 	      break;
 
 	    case IGNORE_SPACE_CHANGE:
 	      /* For -b, advance past any sequence of white space in
 		 line 1 and consider it just one space, or nothing at
 		 all if it is at the end of the line.  */
-	      if (ISSPACE (c1))
+	      if (isspace (c1))
 		{
 		  while (c1 != '\n')
 		    {
 		      c1 = *t1++;
-		      if (! ISSPACE (c1))
+		      if (! isspace (c1))
 			{
 			  --t1;
 			  c1 = ' ';
@@ -362,12 +358,12 @@ lines_differ (char const *s1, char const *s2)
 		}
 
 	      /* Likewise for line 2.  */
-	      if (ISSPACE (c2))
+	      if (isspace (c2))
 		{
 		  while (c2 != '\n')
 		    {
 		      c2 = *t2++;
-		      if (! ISSPACE (c2))
+		      if (! isspace (c2))
 			{
 			  --t2;
 			  c2 = ' ';
@@ -382,15 +378,15 @@ lines_differ (char const *s1, char const *s2)
 		     for equality, go back to the first non-white-space
 		     character in both sides and try again.  */
 		  if (c2 == ' ' && c1 != '\n'
-		      && (unsigned char const *) s1 + 1 < t1
-		      && ISSPACE (t1[-2]))
+		      && s1 + 1 < t1
+		      && isspace ((unsigned char) t1[-2]))
 		    {
 		      --t1;
 		      continue;
 		    }
 		  if (c1 == ' ' && c2 != '\n'
-		      && (unsigned char const *) s2 + 1 < t2
-		      && ISSPACE (t2[-2]))
+		      && s2 + 1 < t2
+		      && isspace ((unsigned char) t2[-2]))
 		    {
 		      --t2;
 		      continue;
@@ -423,7 +419,7 @@ lines_differ (char const *s1, char const *s2)
 			break;
 		    }
 		  if (column != column2)
-		    return 1;
+		    return true;
 		}
 	      break;
 
@@ -435,20 +431,20 @@ lines_differ (char const *s1, char const *s2)
 
 	  if (ignore_case)
 	    {
-	      c1 = TOLOWER (c1);
-	      c2 = TOLOWER (c2);
+	      c1 = tolower (c1);
+	      c2 = tolower (c2);
 	    }
 
 	  if (c1 != c2)
 	    break;
 	}
       if (c1 == '\n')
-	return 0;
+	return false;
 
       column += c1 == '\t' ? tabsize - column % tabsize : 1;
     }
 
-  return 1;
+  return true;
 }
 
 /* Find the consecutive changes at the start of the script START.
@@ -510,7 +506,8 @@ print_script (struct change *script,
 
 /* Print the text of a single line LINE,
    flagging it with the characters in LINE_FLAG (which say whether
-   the line is inserted, deleted, changed, etc.).  */
+   the line is inserted, deleted, changed, etc.).  LINE_FLAG must not
+   end in a blank, unless it is a single blank.  */
 
 void
 print_1_line (char const *line_flag, char const *const *line)
@@ -521,12 +518,25 @@ print_1_line (char const *line_flag, char const *const *line)
 
   /* If -T was specified, use a Tab between the line-flag and the text.
      Otherwise use a Space (as Unix diff does).
-     Print neither space nor tab if line-flags are empty.  */
+     Print neither space nor tab if line-flags are empty.
+     But omit trailing blanks if requested.  */
 
   if (line_flag && *line_flag)
     {
-      flag_format = initial_tab ? "%s\t" : "%s ";
-      fprintf (out, flag_format, line_flag);
+      char const *flag_format_1 = flag_format = initial_tab ? "%s\t" : "%s ";
+      char const *line_flag_1 = line_flag;
+
+      if (suppress_blank_empty && **line == '\n')
+	{
+	  flag_format_1 = "%s";
+
+	  /* This hack to omit trailing blanks takes advantage of the
+	     fact that the only way that LINE_FLAG can end in a blank
+	     is when LINE_FLAG consists of a single blank.  */
+	  line_flag_1 += *line_flag_1 == ' ';
+	}
+
+      fprintf (out, flag_format_1, line_flag_1);
     }
 
   output_1_line (base, limit, flag_format, line_flag);
@@ -545,7 +555,7 @@ output_1_line (char const *base, char const *limit, char const *flag_format,
 	       char const *line_flag)
 {
   if (!expand_tabs)
-    fwrite (base, limit - base, 1, outfile);
+    fwrite (base, sizeof (char), limit - base, outfile);
   else
     {
       register FILE *out = outfile;
@@ -582,8 +592,7 @@ output_1_line (char const *base, char const *limit, char const *flag_format,
 	    break;
 
 	  default:
-	    if (ISPRINT (c))
-	      column++;
+	    column += isprint (c) != 0;
 	    putc (c, out);
 	    break;
 	  }
@@ -606,13 +615,13 @@ translate_line_number (struct file_data const *file, lin i)
 }
 
 /* Translate a line number range.  This is always done for printing,
-   so for convenience translate to long rather than lin, so that the
+   so for convenience translate to long int rather than lin, so that the
    caller can use printf with "%ld" without casting.  */
 
 void
 translate_range (struct file_data const *file,
 		 lin a, lin b,
-		 long *aptr, long *bptr)
+		 long int *aptr, long int *bptr)
 {
   *aptr = translate_line_number (file, a - 1) + 1;
   *bptr = translate_line_number (file, b + 1) - 1;
@@ -627,7 +636,7 @@ translate_range (struct file_data const *file,
 void
 print_number_range (char sepchar, struct file_data *file, lin a, lin b)
 {
-  long trans_a, trans_b;
+  long int trans_a, trans_b;
   translate_range (file, a, b, &trans_a, &trans_b);
 
   /* Note: we can have B < A in the case of a range of no lines.
@@ -662,7 +671,7 @@ analyze_hunk (struct change *hunk,
   lin show_from, show_to;
   lin i;
   bool trivial = ignore_blank_lines || ignore_regexp.fastmap;
-  size_t trivial_length = (int) ignore_blank_lines - 1;
+  size_t trivial_length = ignore_blank_lines - 1;
     /* If 0, ignore zero-length lines;
        if SIZE_MAX, do not ignore lines just because of their length.  */
   bool skip_leading_white_space =
@@ -691,7 +700,7 @@ analyze_hunk (struct change *hunk,
 	  size_t len = newline - line;
 	  char const *p = line;
 	  if (skip_leading_white_space)
-	    while (ISSPACE ((unsigned char) *p) && *p != '\n')
+	    while (isspace ((unsigned char) *p) && *p != '\n')
 	      p++;
 	  if (newline - p != trivial_length
 	      && (! ignore_regexp.fastmap
@@ -706,7 +715,7 @@ analyze_hunk (struct change *hunk,
 	  size_t len = newline - line;
 	  char const *p = line;
 	  if (skip_leading_white_space)
-	    while (ISSPACE ((unsigned char) *p) && *p != '\n')
+	    while (isspace ((unsigned char) *p) && *p != '\n')
 	      p++;
 	  if (newline - p != trivial_length
 	      && (! ignore_regexp.fastmap
@@ -747,17 +756,6 @@ zalloc (size_t size)
   memset (p, 0, size);
   return p;
 }
-
-/* Yield the newly malloc'd pathname
-   of the file in DIR whose filename is FILE.  */
-
-char *
-dir_file_pathname (char const *dir, char const *file)
-{
-  char const *base = base_name (dir);
-  bool omit_slash = !*base || base[strlen (base) - 1] == '/';
-  return concat (dir, "/" + omit_slash, file);
-}
 
 void
 debug_script (struct change *sp)
@@ -766,10 +764,10 @@ debug_script (struct change *sp)
 
   for (; sp; sp = sp->link)
     {
-      long line0 = sp->line0;
-      long line1 = sp->line1;
-      long deleted = sp->deleted;
-      long inserted = sp->inserted;
+      long int line0 = sp->line0;
+      long int line1 = sp->line1;
+      long int deleted = sp->deleted;
+      long int inserted = sp->inserted;
       fprintf (stderr, "%3ld %3ld delete %ld insert %ld\n",
 	       line0, line1, deleted, inserted);
     }
